#include <avr/eeprom.h>
#include <Arduino.h>
#include <OneWire.h>
#include <DallasTemperature.h>

#include "buzzer.h"
#include "DS1340.h"
#include "ht1632.h"
#include "displaytiles.h"

// The DS18B20 Temperature sensor interface
#define ONE_WIRE_BUS 10
static OneWire				oneWire(ONE_WIRE_BUS);
static DallasTemperature	tempSensor(&oneWire);	// Pass our oneWire reference to Dallas Temperature. 

static byte nextColor(byte color)
{
  if (color == ORANGE)
    return GREEN;
  else
    color++;
  return color;
}


/* ------------------------------------------------
 * Time Display class
 * ------------------------------------------------ */
TimeTile::TimeTile()
{
	_color = RED;
	_line = 1;
	_count = 0;
	_state = TIMETILE_STATE_ON;
}


void TimeTile::init() {}


void TimeTile::_show(int h, int m, int s, byte ch, byte cm, byte cs)
{
	static char TwoDigits[] = "%02d";
	char buffer[20];
	
	sprintf(buffer, TwoDigits, h);
	ledMatrixDisplay.print(buffer, _line, ch);
	sprintf(buffer, TwoDigits, m);
	ledMatrixDisplay.print(13, _line, buffer, cm);
	sprintf(buffer, TwoDigits, s);
	ledMatrixDisplay.print(25, _line, buffer, cs, 3);
}


void TimeTile::show()
{
	DateAndTime *pdt;
	int h, m, s;
	
	pdt = rtcModule.getTime(NULL);
	
	h = pdt->Hour;
	m = pdt->Minute;
	s = pdt->Second;
	switch(_state)
	{
		case TIMETILE_STATE_ON:
			_show(h, m, s, _color, _color, _color);
			break;
		case TIMETILE_STATE_SETH:
			_show(h, m, s, RED, GREEN, GREEN);
			break;
		case TIMETILE_STATE_SETM:
			_show(h, m, s, GREEN, RED, GREEN);
			break;
		case TIMETILE_STATE_SETS:
			_show(h, m, s, GREEN, GREEN, RED);
			break;
	}
}

void TimeTile::tick()
{
	if (_state != TIMETILE_STATE_ON)
		return;
	_count++;
	_count %= 9; // Once in 9 seconds
	if (_count == 0)
		_color = nextColor(_color);
}


bool TimeTile::sendKey(int keyCode)
{
	DateAndTime *pdt;
	bool rv = false;
	bool saveToRTC = false;
	
	pdt = rtcModule.getTime(NULL);

	switch(_state)
	{
		case TIMETILE_STATE_ON:
			if (keyCode == KEYCODE_SET)
			{
				_state = TIMETILE_STATE_SETH;
				rv = true;
			}
			else
				rv = false;
			break;
		case TIMETILE_STATE_SETH:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					pdt->Hour++;
					if (pdt->Hour > 23)
						pdt->Hour = 0;
					saveToRTC = true;					
					break;
				case KEYCODE_MINUS:
					if (pdt->Hour > 0)
						pdt->Hour--;
					else
						pdt->Hour = 23;
					saveToRTC = true;
					break;
				case KEYCODE_SET:
					_state = TIMETILE_STATE_SETM;
					break;
			}
			rv = true;
			break;
		case TIMETILE_STATE_SETM:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					pdt->Minute++;
					if (pdt->Minute > 59)
						pdt->Minute = 0;
					saveToRTC = true;
					break;
				case KEYCODE_MINUS:
					if (pdt->Minute > 0)
						pdt->Minute--;
					else
						pdt->Minute = 59;
					saveToRTC = true;
					break;
				case KEYCODE_SET:
					_state = TIMETILE_STATE_SETS;
					break;
			}
			rv = true;
			break;
		case TIMETILE_STATE_SETS:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
				case KEYCODE_MINUS:
					pdt->Second = 0;
					saveToRTC = true;
					rv = true;
					break;
				case KEYCODE_SET:
					_state = TIMETILE_STATE_ON;
					break;
			}
			rv = false;
			break;
	}
	if (saveToRTC)
		rtcModule.setRTCTime(NULL);
	return rv;
}

/* ------------------------------------------------
 * Date Display class
 * ------------------------------------------------ */
DateTile::DateTile()
{
	_color = GREEN;
	_line = 2;
	_count = 0;
	_state = DATETILE_STATE_ON;
}


void DateTile::init() {}


void DateTile::show()
{
	#define _DATETILEMODE_DDDdd		0
	#define _DATETILEMODE_MMMyy		1
	#define _DATETILEMODE_ddmmyyyy 	2
	#define _DATETILEMODE_MAX 		2
	static const char monthNames[][4] = 
		{ "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC" };
	static const char dowNames[][4] = { "SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT" };
	static const char twoDecDigits[] = "%02d";	
	static unsigned long _last = 0UL;		// Last time we flipped the tile content
    static byte dateTileMode = _DATETILEMODE_DDDdd; // The current date tile mode/format
	static int lastVal = -1;	// When in set-date saves the last value in edit
	
	DateAndTime *pdt;
	char buffer[20];
	int day, month, year, dow;
	unsigned long _now = millis();
	byte colorDay = GREEN;
	byte colorMonth = GREEN;
	byte colorYear = GREEN;
	bool needRefresh = true;	// If to refresh the entire tile or not

	if (_state == DATETILE_STATE_ON)
	{
		if ((_now - _last) < 2000UL) // No need to refresh display if less than 2 seconds ago
			return;
		if (_last == 0UL)
			dateTileMode = _DATETILEMODE_DDDdd;	// The fist tile format to use
		else
		{
			// Circular loop through the different formats for this tile
			dateTileMode++;
			dateTileMode %= (1 + _DATETILEMODE_MAX);
		}
	}
	else
		dateTileMode = _DATETILEMODE_ddmmyyyy;
    _last = _now;

	pdt = rtcModule.getTime(NULL);
	day = pdt->Date;
	month = pdt->Month;
	year = pdt->Year;
	dow = pdt->Day;
	
	switch (dateTileMode)
	{
		case _DATETILEMODE_DDDdd:
			ledMatrixDisplay.clearLine(_line);
			ledMatrixDisplay.print(dowNames[dow-1], _line, _color);
			sprintf(buffer, twoDecDigits, day);
			ledMatrixDisplay.print(20, _line, buffer, _color, 5);
			break;
		case _DATETILEMODE_MMMyy:
			ledMatrixDisplay.clearLine(_line);
			ledMatrixDisplay.print(monthNames[month-1], _line, _color);
			sprintf(buffer, twoDecDigits, year);
			ledMatrixDisplay.print(20, _line, buffer, _color, 5);
			break;
		case _DATETILEMODE_ddmmyyyy:
			switch (_state)
			{
				case DATETILE_STATE_ON:
					colorDay = _color;
					colorMonth = nextColor(_color);
					colorYear = nextColor(colorMonth);
					break;
				case DATETILE_STATE_SETD:
					colorDay = RED;
					_color = colorMonth = colorYear = GREEN;
					needRefresh = (day != lastVal); // if nothing changed, don't refresh
					lastVal = day;
					break;
				case DATETILE_STATE_SETM:
					_color = colorDay = colorYear = GREEN;
					colorMonth = RED;
					needRefresh = (month != lastVal); // if nothing changed, don't refresh
					lastVal = month;
					break;
				case DATETILE_STATE_SETY:
					_color = colorDay = colorMonth = GREEN;
					colorYear = RED;
					needRefresh = (lastVal != year);
					lastVal = year;
					break;
			}
			if (needRefresh)// Refresh only when needed. If this is not done, during
							// set date mode the display will flicker. In standard
							// display mode we refresh only every 2 seconds
				ledMatrixDisplay.clearLine(_line);
			sprintf(buffer, twoDecDigits, day);
			ledMatrixDisplay.print(0, _line, buffer, colorDay, 3);
			sprintf(buffer, twoDecDigits, month);
			ledMatrixDisplay.print(8, _line, buffer, colorMonth, 3);
			sprintf(buffer, "%4d", year+2000);
			ledMatrixDisplay.print(17, _line, buffer, colorYear, 3);
			//ledMatrixDisplay.plot(7, 15,  _color);
			//ledMatrixDisplay.plot(15, 15, _color);
			break;
	}
}

void DateTile::tick()
{
	if (_state != DATETILE_STATE_ON)
		return;
	_count++;
	_count %= 9; // Once in 9 seconds
	if (_count == 0)
		_color = nextColor(_color);
}


bool DateTile::sendKey(int keyCode)
{
	DateAndTime *pdt;
	bool rv = false;
	bool saveToRTC = false;
	
	pdt = rtcModule.getTime(NULL);

	switch(_state)
	{
		case DATETILE_STATE_ON:
			if (keyCode == KEYCODE_SET)
			{
				_state = TIMETILE_STATE_SETH;
				rv = true;
			}
			else
				rv = false;
			break;
		case DATETILE_STATE_SETD:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					pdt->Date++;
					if (pdt->Date > 31)
						pdt->Date = 1;
					saveToRTC = true;					
					break;
				case KEYCODE_MINUS:
					if (pdt->Date > 1)
						pdt->Date--;
					else
						pdt->Date = 31;
					saveToRTC = true;
					break;
				case KEYCODE_SET:
					_state = DATETILE_STATE_SETM;
					break;
			}
			rv = true;
			break;
		case DATETILE_STATE_SETM:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					pdt->Month++;
					if (pdt->Month > 12)
						pdt->Month = 1;
					saveToRTC = true;
					break;
				case KEYCODE_MINUS:
					if (pdt->Month > 1)
						pdt->Month--;
					else
						pdt->Month = 12;
					saveToRTC = true;
					break;
				case KEYCODE_SET:
					_state = DATETILE_STATE_SETY;
					break;
			}
			rv = true;
			break;
		case DATETILE_STATE_SETY:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					pdt->Year++;
					saveToRTC = true;
					rv = true;
					break;
				case KEYCODE_MINUS:
					pdt->Year--;
					saveToRTC = true;
					rv = true;
					break;
				case KEYCODE_SET:
					_state = DATETILE_STATE_ON;
					break;
			}
			rv = false;
			break;
	}
	if (saveToRTC)
		rtcModule.setRTCTime(NULL);
	return rv;
}

void DateTile::startSetting()
{
	_state = DATETILE_STATE_SETD;
}


/* ------------------------------------------------
 * Temperature Display class
 * ------------------------------------------------ */
TempTile::TempTile()
{
	_line = 2;
}


void TempTile::init()
{
	int num;

    Serial.println("Searching OneWire bus");
	tempSensor.begin();
	num = tempSensor.getDeviceCount();
	Serial.print("Found "); Serial.println(num, DEC);
}

void TempTile::show()
{
	static unsigned long _last = 0UL;
	unsigned long _now = millis();
	char buff[10];
	int color;
        int temp;
        float f;

	if ((_now - _last) < 5000)
		return;
        _last = _now;

	// Read temperature
	tempSensor.requestTemperatures();
	f = tempSensor.getTempCByIndex(0);
        temp = (int)(10.0F * f);
	ledMatrixDisplay.clearLine(_line);
	sprintf(buff, "%2d", (temp/10));
	if (temp < 225 )
		color = GREEN;
	else
		if (temp > 275)
			color = RED;
		else
			color = ORANGE;
        temp %= 10;
	ledMatrixDisplay.print(buff, _line, color);
	sprintf(buff, "%d", temp);
    ledMatrixDisplay.print(14, _line, buff, color);
	ledMatrixDisplay.plot(12, 15, color); // The decimal point
	ledMatrixDisplay.plot(20, 9, color); // The degrees "symbol"
    ledMatrixDisplay.print(22, _line, "C", color);
}


void TempTile::tick() { }



/* ------------------------------------------------
 * Countdown timer display class
 * ------------------------------------------------ */
CountdownTile::CountdownTile()
{
	_line = 2;
}

void CountdownTile::_ReadFromEprom()
{
	byte val;
	
	val = eeprom_read_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR);
	if (val != COUNTDOWN_EEPROM_SIGN)
	{
		Serial.print("EEPROM is incorrect");
		minutes = 5;
		seconds = 0;
	}
	minutes = eeprom_read_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR+1);
	seconds = eeprom_read_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR+2);
}


void CountdownTile::_WriteToEprom(byte m, byte s)
{
	eeprom_write_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR, COUNTDOWN_EEPROM_SIGN);
	eeprom_write_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR+1, (uint8_t)m);
	eeprom_write_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR+2, (uint8_t)s);
}


void CountdownTile::init()
{
	byte val;
	
	val = eeprom_read_byte((unsigned char *)COUNTDOWN_EEPROM_ADDR);
	if (val != COUNTDOWN_EEPROM_SIGN)
	{
		// Initialize EEPROM with 5 minutes
		Serial.print("Initializing EEPROM with 5min countdown");
		_WriteToEprom(5, 0);
	}
      _ReadFromEprom();
}

void CountdownTile::_startCountDown()
{
	millisStart = millis();
}


void CountdownTile::_doCountDown()
{
	unsigned long _now = millis();
	
	if (_now - millisStart > 1000)
	{
		currSeconds--;
		millisStart = _now;
	}
}


void CountdownTile::_show(byte mColor, byte sColor)
{
	char buff[20];
	int min, sec;
	byte aColor = (_state == COUNTDOWN_STATE_RUNNING) ? RED : ORANGE;
	byte dColor;
	static char twoDigits[] = "%02d";	
	
	if (mColor == sColor)
	{
		min = currSeconds / 60;
		sec = currSeconds % 60;
		dColor = mColor;
	}
	else
	{
		min = minutes;
		sec = seconds;
		dColor = GREEN;
	}
	sprintf(buff, twoDigits, min);
	ledMatrixDisplay.print(buff, _line, mColor);
	sprintf(buff, twoDigits, sec);
	ledMatrixDisplay.print(15, _line, buff, sColor);
	// The two dots between minutes and seconds
	ledMatrixDisplay.plot(13, 10, dColor);
	ledMatrixDisplay.plot(13, 13, dColor);
    // the down arrow symbol
	ledMatrixDisplay.print(27, _line, "d", aColor); // "d" character is mapped to down arrow in font
}


void CountdownTile::show()
{
	bool stay = true;
	byte color;

	do
	{
		switch(_state)
		{
			case COUNTDOWN_STATE_OFF:
				stay = false;
				return;
			case COUNTDOWN_STATE_ON:
				ledMatrixDisplay.clearLine(_line);
			case COUNTDOWN_STATE_WAITING:
				_state = COUNTDOWN_STATE_WAITING;
				currSeconds = 60 * minutes + seconds;
				_show(GREEN, GREEN);
				stay = false;
				break;
			case COUNTDOWN_STATE_RUNNING:
				_doCountDown();
				if (currSeconds == 0)
				{
					// Stop the countdown, play a tune and go back to ON state
					_state = COUNTDOWN_STATE_ON;
					buzzer.playEndTune();
					break;
				}
				stay = false;
				if (currSeconds > 60)
					color = GREEN;
				else
					if (currSeconds < 10)
						color = RED;
					else
						color = ORANGE;
				_show(color, color);
				break;
			case COUNTDOWN_STATE_SETMIN:
				stay = false;
				_show(RED, GREEN);
				break;
			case COUNTDOWN_STATE_SETSEC:
				stay = false;
				_show(GREEN, RED);
				break;
		}
	} while (stay);
}


void CountdownTile::tick() { }


bool CountdownTile::sendKey(int keyCode)
{
	switch(_state)
	{
		case COUNTDOWN_STATE_OFF:
			if (keyCode == KEYCODE_TIMER)
			{
				_state = COUNTDOWN_STATE_ON;
				return true;
			}
			return false;
                case COUNTDOWN_STATE_ON:
		case COUNTDOWN_STATE_WAITING:
			switch(keyCode)
			{
				case KEYCODE_SET:
					_state = COUNTDOWN_STATE_SETMIN;
					return true;
				case KEYCODE_TIMER:
					_state = COUNTDOWN_STATE_RUNNING;
					_startCountDown();
					return true;
			}
			return true;
		case COUNTDOWN_STATE_RUNNING:
			if (keyCode == KEYCODE_TIMER) // if timer pressed during run, leave countdown
			{
				_state = COUNTDOWN_STATE_OFF;
				return false;
			}
			return true;
		case COUNTDOWN_STATE_SETMIN:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					minutes++;
					minutes %= 99;
					break;
				case KEYCODE_MINUS:
					minutes--;
					if (minutes >= 99) minutes = 0;
					break;
				case KEYCODE_SET:
					_state = COUNTDOWN_STATE_SETSEC;
					break;
			}
			return true;
		case COUNTDOWN_STATE_SETSEC:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					seconds++;
					seconds %= 60;
					break;
				case KEYCODE_MINUS:
					seconds--;
					if (seconds >= 60) seconds = 0;
					break;
				case KEYCODE_SET:
					_WriteToEprom(minutes, seconds);
					_state = COUNTDOWN_STATE_ON;
					break;
			}
			return true;
	}
	return false;
}



/* ------------------------------------------------
 * The display manager class
 * ------------------------------------------------ */
DisplayManager::DisplayManager() { intensity = 15;}


void DisplayManager::init() 
{
	time.init();
	date.init();
	temperature.init();
	countdown.init();
	ledMatrixDisplay.setIntensity(intensity);
}


void DisplayManager::show()
{
	static unsigned long _last = 0UL;
	unsigned long _now;
	static bool showTemp = false;
    bool showTimer = false;
	unsigned int swapTime;

	_now = millis();
        swapTime = (showTemp) ? 3000U : (3U * 2000U); // Temperature for 3 seconds, date has three formats, 2 seconds each
	if (_last == 0UL)
	{
		showTemp = false;
		_last = _now;
	}
	else
		if ( (_now - _last) > swapTime )
		{
			showTemp ^= true;	// Swap date and temperature
			_last = _now;
		}
        if (owner == KEYOWNER_TIMER)
        {
          showTimer = true;
          showTemp = false;
        }
        else
  	  if (owner == KEYOWNER_DATE)	// Don't allow temperature tile if setting date
	  {
		showTemp = false;
		_last = _now;		// To ensure no flickering to temp once done editing date
	  }

	time.show();
	if (showTemp)
		temperature.show();
	else
		if (showTimer)
			countdown.show();
		else
			date.show();
}


void DisplayManager::tick()
{
	time.tick();
	date.tick();
	temperature.tick();
	countdown.tick();
}


void DisplayManager::sendKey(int keyCode)
{
	switch(owner)
	{
		case KEYOWNER_MYSELF:
			switch(keyCode)
			{
				case KEYCODE_PLUS:
					if (intensity < 15) intensity++;
					ledMatrixDisplay.setIntensity(intensity);
					break;
				case KEYCODE_MINUS:
					if (intensity > 0) intensity--;
					ledMatrixDisplay.setIntensity(intensity);
					break;
				default:
					// Seatch new key owner
					if (time.sendKey(keyCode))
						owner = KEYOWNER_TIME;
					else
						if (date.sendKey(keyCode))
							owner = KEYOWNER_DATE;
						else
							if (countdown.sendKey(keyCode))
								owner = KEYOWNER_TIMER;
							else
								owner = KEYOWNER_MYSELF;
					break;
			}
			break;
		case KEYOWNER_TIME:
			if (time.sendKey(keyCode) == false)
			{
				// When the TIME tile is no longer the owner of the keys,
				// we are moving to date ownership as the SET button
				// is going through hour->minute->second|->day->month->year
				// when in SET is pressed
				owner = KEYOWNER_DATE;
                                date.sendKey(keyCode);
			}
			break;
		case KEYOWNER_DATE:
			if (date.sendKey(keyCode) == false)
				owner = KEYOWNER_MYSELF;
			break;
		case KEYOWNER_TIMER:
			if (countdown.sendKey(keyCode) == false)
				owner = KEYOWNER_MYSELF;
			break;
	}
}



DisplayManager	displayManager;